clc, clear, close all
%% Path-Planning on a BINN Landscape Model for an ASV
    % Shell algorithm
    % Targets are minima in the BINN
    % LOS guidance law

%% Plotting preferences and choice of heuristic function
plot_search = 0; % plot hybrid A* search tree: 1 = on, 0 = off
plot_path   = 0; % plot hybrid A* path at each iteration: 1 = on, 0 = off
expansion   = 0; % 0 = straight-line, 1 = kinematics-based
heuristics  = 0; % 0 = Euclidean distance, 1 = adjusted for neural activity

%% Angle conversion factors
rad2deg = 180/pi; % radians to degrees
deg2rad = pi/180; % degrees to radians

%% Configuration
% Workspace parameters
Lx = 42; % Length in x-direction (north)
Ly = Lx; % Length in y-direction (east)

Nx = 14; dX = Lx/Nx; % Grid partitioning (regular)
Ny = Nx; dY = Ly/Ny;
N  = Nx*Ny;          % Total number of grid cells/neurons

[X,Y] = meshgrid(dX/2:dX:Lx,dY/2:dY:Ly);                 
x     = reshape(X,[N,1]);
y     = reshape(Y,[N,1]);

% Simulation parameters
h       = 0.01;             % [s] Sample time
N_inner = 1000;             % For pre-allocation in inner loop
N_outer = 100;              % For pre-allocation in outer loop

% ASV parameters
L      = 1.105;             % Vehicle length
Delta  = L;                 % LOS lookahead distance
Kp     = 0.5;               % Proportional gain (course controller) 
Uref   = .2;                % [m/s] Speed of vehicle (forward)
Rmin   = 3/2 *L;            % [m] Minimum turn radius
r_max  = Uref/Rmin;         % [rad/s] Maximum rotation rate (14deg/s for HUGIN with u=2m/s in Wiig et al. 2017)
idx0   = Ny;                % Initial position (index in vector grid)
psi0   = 0*deg2rad;         % Initial heading (towards north)
r0     = min(dX,dY)/6;      % Radius of acceptance (defining when the waypoint is regarded reached by the vessel)

% Parameters for COLAV compliant with COLREGs
ENCOUNTER_STATUS = 'SO';    % OTp/s = overtaking (port/starboard), HO = head on, GW = give way, SO = stand on
rd   = 20;                  % Radius of the detection region
sd_L = 10*L;                % large safe distance
sd_M = 5*L;                 % medium safe distance
sd_S = 2*L;                 % small safe distance

% Neural Network parameters
E    = 10;                  % Constant for external input
d0   = max(dX,dY);          % Radius of the receptive field R (defining the neighboring neurons)
a    = 10;                  % Passive decay rate (shunting eq.)
bu   = 1.0;                 % Upper bound of neural activity (shunting eq.)            
bl   = 1.0;                 % Lower bound of neural activity (shunting eq.)
mu   = 1;                   % Constant defining connection weight between neurons
e_ss = 1e-10;               % Condition for steady-state neural activity 

%UPS  = zeros(size(X));      % Grid of external inputs (alt. UPS = zeros(Nx,Ny))
ups  = zeros(size(x));      % External inputs
z    = zeros(size(x));      % Neural activities

% Determining neighbouring neurons A(i,j) and connection weights W(i,j)
A = zeros(N,N);             % Adjacency matrix
W = zeros(N,N);             % Connection weights between neurons
for i = 1:N
    for j = 1:N
        dist = max(abs(x(j)-x(i)),abs(y(j)-y(i)));
%        dist = sqrt((x(j)-x(i))^2+(y(j)-y(i))^2);
        if (i ~= j) && (dist <= d0)
            A(i,j) = 1;
            W(i,j) = mu/dist;
        end
    end
end

%% Obstacles and Targets
% Targets
TargetIdx = [Nx-1, 1]; %[floor(Nx/2), floor(Ny/2)]; % Single target in upper left corner  
[Nt,~] = size(TargetIdx);   % Number of targets
idx_t = (TargetIdx(:,1)-1).*Ny ...
    + TargetIdx(:,2);       % Indices of targets in vector grid          
ups(idx_t) = -E;            % External input to BINN model

% Static obstacles
StaticObIdx = [6,5; 7,3; 7,4; 7,5; 7,6; 8,4; 8,5; 8,6; 8,7; 12,6; 13,6; 14,6];%[6:8;13:15]';
if ~isempty(StaticObIdx)
    idx_so = (StaticObIdx(:,1)-1).*Ny ...
        + StaticObIdx(:,2); % Indices of static obstacles in vector grid       
    ups(idx_so) = E;        % External input to BINN model 
else
    idx_so = [];
end

% Moving obstacle
theta   = linspace(0, 2*pi);
r_obj   = 1;                              % [m] radius of object
obj_def = r_obj*[cos(theta); sin(theta)]; % object definition (circle)            
x_obj = 27;%Lx/5*(1+2*rand);              % [m] initial position
y_obj = 39;%Ly/5*(1+2*rand);             
p_obj = [[x_obj,y_obj]',...               % object tracks
    zeros(2,N_outer*N_inner)];                                     
U_obj = 0.9*Uref;                         % [m/s] velocity 
d_obj = -3*pi/4;%2*pi*rand;                % [rad] initial direction 

% Assume that the initial vessel position & orientation is feasible
idx_av = [];
avoidIdx = [];
ObstacleIdx = [StaticObIdx; avoidIdx];

%UPS  = reshape(ups,[Ny,Nx]);
%figure(5); surf(Y,X,UPS); xlabel('y'); ylabel('x');

%% Preliminaries for plotting of the path
plot_prelims
figure(1); clf; hold on
if ~isempty(StaticObIdx)
    plot(y(idx_so),x(idx_so),'ks','MarkerFaceColor','k',...
        'DisplayName','static obstacles'); % Static obstacles
end
[xGrid,yGrid] = meshgrid(0:dX:Lx,0:dY:Ly); % Grid
plot(yGrid,xGrid,':','Color',gray,'HandleVisibility','off');
plot(yGrid',xGrid',':','Color',gray,'HandleVisibility','off');
xlabel('y [m]'); xticks(dY:2*dY:Ly);
ylabel('x [m]'); yticks(dX:2*dX:Lx);
axis([0 Ly 0 Lx]), axis equal tight manual, box on, hold off;

%% Simulating the motion of a vehicle
psi_path = psi0;               % Current path orientation
path     = [x(idx0),y(idx0);   % Initial WP
            zeros(N_outer,2)]; % Pre-allocate memory for future WPs
idx      = idx0;               % Index to current node/cell

p_asv    = [x(idx0),y(idx0)]'; % Current vehicle position: [x y]'
psi_asv  = psi0;               % Current vehicle heading angle
x_asv    = [p_asv', psi_asv;   % Vehicle state/tracks
            zeros(N_outer*N_inner,3)];  

kk_long  = eye(N_outer,1);     % # iterations for the ASV to reach each WP 
kk       = 2;                  % total # iterations in outer and ASV loop 

% Outer loop. Termination criteria: All targets have been visited
targetsReached = 0;            % no targets are reached yet
j = 1;                         % iterations in outer loop
t_cpu = zeros(N_outer,1);      % [s] Time for path search
while targetsReached < Nt 
    
    ups(idx) = 0.0; % Zero external input to visited node
    z(idx)   = 0.0; % Zero neural activity at visited node
    
    % Evolving NN dynamics
    ss = 0; % steady-state is false
    while(~ss || max(abs(z-z_old)) > e_ss)
        ss = ss + 1;           % # iterations needed to reach steady-state
        z_old = z;
        z = shuntingEq(N, h, ups, A, W, z_old, a, bu, bl);
    end
    
    q = [x, y, z];             % Update neural activity
    Z  = reshape(z,[Ny,Nx]);   % Updata neural activity landscape
%     figure(4); surf(Y,X,Z); xlabel('y'); ylabel('x');
    
    % Search algorithm to find next waypoint
    tic;
    switch expansion
        case 1
            [psi_path, path(j+1,:), idx, path_plan] =... 
                hybridAstarBINN_kinematics(Z, path(j,:), psi_path, Uref,...
                r_max, Nx, Ny, dX, dY, Lx, Ly, x, y, plot_search,...
                plot_path, heuristics);
        otherwise
            [psi_path, path(j+1,:), idx, path_plan] = hybridAstarBINN(Z,...
                path(j,:), psi_path, Nx, Ny, dX, dY, Lx, Ly, x, y,...
                plot_search, plot_path, heuristics);
    end
    t_cpu(j) = toc;
    
    % New subpath rotation
    Rt_path = [ cos(psi_path) sin(psi_path); 
              -sin(psi_path) cos(psi_path)];
      
    % Evolving vehicle dynamics
    k = 1;
    while norm(p_asv - path(j+1,:)') > r0             
        R0          = [cos(psi_asv) -sin(psi_asv); 
                       sin(psi_asv)  cos(psi_asv)];
        q_err       = Rt_path*(p_asv-path(j,:)'); % Pos. error from path
        psi_los     = ssa(psi_path - atan2(q_err(2),Delta),'rad');
        q_err_dot   = Rt_path*R0*[Uref,0]';
        psi_los_dot = -(Delta*q_err_dot(2))/(Delta^2+q_err_dot(2)^2);

        p_asv        = p_asv + h*(R0*[Uref,0]');
        psi_asv      = ssa(psi_asv + ...
            h*(-Kp*ssa(psi_asv - psi_los,'rad') + psi_los_dot),'rad');        
        x_asv(kk,:)  = [p_asv',psi_asv];
        
        kk = kk + 1;
        k  = k + 1;    
    end 
    if sum(idx == idx_t) % Check whether a target is reached
        targetsReached = targetsReached + 1;
    end
    
    % Evolving moving object dynamics
    if path(j+1,:) == path(j,:)  % vessel in DP mode
        k = 1000;                % k*h = 10 [s]
        x_asv(kk:kk+k,:) = repmat([p_asv',psi_asv],k+1,1);
        kk = kk + k + 1;
    end
    for k2 = kk_long(j)+1:kk-1
        dp_obj = h*U_obj*[cos(d_obj), sin(d_obj)]';  % progression
        p_obj(:,k2) = p_obj(:,k2-1) + dp_obj;        % object tracks
        if p_obj(1,k2)<0 || p_obj(1,k2)>Lx || p_obj(2,k2)<0 || p_obj(2,k2)>Ly
            p_obj(:,k2) = p_obj(:,k2-1); 
        end % force the object to stay within the workspace
        %d_obj  = d_obj + 2*deg2rad*(0.5-rand); % direction +/-1.0 degrees
    end
    kk_long(j+1) = kk - 1;
    
    % Plotting
    figure(1); hold on
    delete(findall(figure(1),{'type','patch','-and',{'facecolor',blue,...
        '-or','facecolor',orange}},'-or',{'type','text','color',blue},...
        '-or','type','textboxshape','-or','color',darkgray,...
        '-or',{'type','line','-and','color','g'})); % Delete data from last iteration
    Rd = rd * [cos(theta); sin(theta)] + p_asv; % Current detection region
    fill(Rd(2,:),Rd(1,:),blue,'LineStyle','none','FaceAlpha',.05,...
        'DisplayName','detection region');
    plot(path_plan(:,2),path_plan(:,1),'.',...%'LineWidth',1.5,...
        'MarkerSize',10,'Color',darkgray,'DisplayName','planned WPs'); % Planned path
    plot(y(idx_t),x(idx_t),'p','Color',red,'MarkerFaceColor',red,...
        'MarkerSize',8,'DisplayName','target'); % Targets
    fill(obj_def(2,:)+p_obj(2,kk-1),obj_def(1,:)+p_obj(1,kk-1),... % Current object position
        orange,'FaceAlpha',.2,'EdgeColor',orange,'HandleVisibility','off'); 
    plot(p_obj(2,1:kk-1),p_obj(1,1:kk-1),'-.','Color',orange,...
        'LineWidth',1,'DisplayName','object tracks') % Moving object tracks
    plot(path(1:j+1,2),path(1:j+1,1),'.-','LineWidth',2,...
        'MarkerSize',10,'Color',green,'DisplayName','WP path'); % Path
    plot(x_asv(1:kk-1,2),x_asv(1:kk-1,1),'-k',...
        'DisplayName','vessel tracks'); % Vessel tracks
    for i = j:j+1 % Ships
        R0 = [cos(x_asv(kk_long(i),3)) -sin(x_asv(kk_long(i),3)); 
              sin(x_asv(kk_long(i),3))  cos(x_asv(kk_long(i),3))];
        ship = R0*[6*L/9 5*L/9 3*L/9 -5*L/9 -5*L/9 3*L/9 5*L/9 6*L/9; 
            0 1*L/9 2*L/9 2*L/9 -2*L/9 -2*L/9 -1*L/9 0];
        plot(x_asv(kk_long(i),2)+ship(2,:),x_asv(kk_long(i),1)+ship(1,:),...
            'k','HandleVisibility','off');
    end
    hold off
    pause(0.05);
    
    % COLAV compliant with COLREGs
%     ongoing_maneuver = false;
    dir_obj = atan2(p_obj(2,kk-1)-p_obj(2,kk_long(j)),...
                     p_obj(1,kk-1)-p_obj(1,kk_long(j))); % [rad] object direction at each step
    ups(setdiff(idx_av,idx_so)) = 0;    % Clear old avoidance region
    [dp_rec,dp] = relativeDistance(p_asv, psi_asv, p_obj(:,kk-1)-p_asv);
    if(norm(p_obj(:,kk-1)-p_asv) < rd && dp(1) > 0)% && abs(dp(2)) < sd_L)% && ~ongoing_maneuver) % If obstacle is within detection region
%         ongoing_maneuver = true;
        psi_rel = ssa(dir_obj-psi_asv,'rad');
        Robj = [cos(dir_obj) -sin(dir_obj); sin(dir_obj) cos(dir_obj)];
        if pi-abs(psi_rel) < pi/8 && abs(dp(2)) < sd_M %&& norm(dp) < sd_L 
           ENCOUNTER_STATUS = 'HO'; % Alter to starboard: Define an area covering the object + margin to starboard and extended at the port side
           p_ahead = max(dp(1)-4*L,2*r_obj+sd_S);
           avoidA0 = [p_ahead,p_ahead,-2*r_obj,-2*r_obj; % area as if the obstacle was in position (0,0) headed towards the north
               Ly,-(2*r_obj+sd_S),-(2*r_obj+sd_S),Ly];
        elseif(psi_rel >= -7*pi/8 && psi_rel < -3*pi/8)
           ENCOUNTER_STATUS = 'GW'; % Pass behind 
           avoidA0 = [Lx,Lx,-2*r_obj,-2*r_obj; % area as if the obstacle was in position (0,0) headed towards the north
               (2*r_obj+sd_M),-(2*r_obj+sd_M),-(2*r_obj+sd_M),(2*r_obj+sd_M)];
        elseif ((psi_rel >= -3*pi/8 && psi_rel < 0) && dp(2) > -sd_S)
            ENCOUNTER_STATUS = 'OTp'; % Alter to starboard
            p_ahead = k*h*U_obj; %-2*r_obj;
            avoidA0 = [p_ahead+sd_L,p_ahead+sd_L,p_ahead,p_ahead; % area as if the obstacle was in position (0,0) headed towards the north
                (2*r_obj+sd_S),-Ly,-Ly,(2*r_obj+sd_S)];
        elseif ((psi_rel >= 0 && psi_rel < 3*pi/8) && dp(2) < sd_S)
            ENCOUNTER_STATUS = 'OTs'; % Alter to port
            p_ahead = k/2 *h*U_obj; %-2*r_obj;
            avoidA0 = [p_ahead+sd_L,p_ahead+sd_L,p_ahead,p_ahead; % area as if the obstacle was in position (0,0) headed towards the north
                Ly,-(2*r_obj+sd_S),-(2*r_obj+sd_S),Ly];
        else 
           ENCOUNTER_STATUS = 'SO'; % Keep course and speed
           avoidA0 = [sd_L,sd_L,-2*r_obj,-2*r_obj; % area as if the obstacle was in position (0,0) headed towards the north
               (2*r_obj+sd_S),-(2*r_obj+sd_S),-(2*r_obj+sd_S),(2*r_obj+sd_S)];
        end
        avoidArot = p_obj(:,kk-1) + Robj*(r_obj*[-1,-1,1,1;-1,1,1,-1] +...
               avoidA0); % rotate avoidance area
        avoidA = [avoidArot,avoidArot(:,1)];
        in = inpolygon(x,y,avoidA(1,:),avoidA(2,:));
        avoidIdx = [ceil(x(in)/dX) ceil(y(in)/dY)];
        idx_av = (avoidIdx(:,1)-1).*Ny + avoidIdx(:,2); % Avoidance region indices in vector grid
        ups(idx_av) = E;
        if ~sum(idx_av==idx_t) % Target is not covered by avoidance region
            ups(idx_t) = -E;
        end
%     elseif ongoing_maneuver
%         % if boats have passed each others, change ongoing_maneuver to false
    else
        ENCOUNTER_STATUS = 'SO'; % Keep course and speed
        idx_av = [];
        avoidIdx = [];
    end
    ObstacleIdx = [StaticObIdx; avoidIdx];
    figure(1); hold on; 
    %plot(dp_rec(2,:),dp_rec(1,:),'g-','DisplayName','relative distance'); % Relative distance
    if ~isempty(idx_av) % Avoidance region  
        fill(avoidA(2,:),avoidA(1,:),orange,'FaceAlpha',.1,...
            'Linestyle','none','DisplayName','avoidance region');
    end
    annotation(figure(1),'textbox',[.565 .823 .1 .1],'String',... 
    ['Encounter status:',' ',ENCOUNTER_STATUS],...
    'FitBoxToText','on','LineStyle','none'); % Type of encounter
%     if j<=1
%         legend('show','Location','southwest','AutoUpdate','off');
%     end
    hold off;
    
    j = j + 1;
end

% Path length and CPU time
t_cpu = t_cpu(1:j-1);
path_length = sum(sqrt(sum(diff(path(1:j,:)).^2,2)));

function [dp_rec1,dp1] = relativeDistance(p1, psi1, dp)
    % Function calculating the distances between two objects relative to 
    % the body axes of object 1 
    % Input:
    %   p1   - position of object 1
    %   psi1 - heading of object 1
    %   dp   - distance between the objects relative to the NE-axes
    % Output: 
    %   dp1     - distance between the objects relative to object 1
    %   dp_rec1 - a rectangle with sides equal to dp1
    %
    % Caroline Fleischer, 05.04.2020
    
    rec0 = [0 1 1 0 0;
            0 0 1 1 0];
    R1 = [cos(psi1) -sin(psi1);
          sin(psi1)  cos(psi1)];
    dp_rec0_rot = R1'*dp.*rec0;
    dp_rec1 = p1 + R1*dp_rec0_rot;
    dp1 = R1'*dp; %abs(R1'*dp)
    
end