function [u, alpha, omega, tau_out] = ctrlAlloc(tau)

    % Mapping of desired force (tau) to input to the actuators (u)
    % Input:  3DOF desired forces and moment in body-frame, tau = [X Y N]'
    % Output: input signals to the actuators, u = [u_BT u_VSP1 u_VSP2]';
    %         angles of the two VSPs, alpha = [alpha_VSP1 alpha_VSP2]';
    %         angular rates of the VSPs, omega = [omega_VSP1 omega_VSP2]'
    %         applied force, tau_out = [X_out Y_out Z_out]' 
    %
    % Author:   Caroline Fleischer, Elias Gauslaa, Jakob Jensen
    % Date:     20.02.2020
    % Modified: Caroline Fleischer, 24.04.2020

    %%%% Copy to init file %%%%
    ly_VSP = 0.055;     % y_position of VSP2
    lx_VSP = 0.4574;    % x-position of the VSPs
    lx_BT = 0.3875;     % x-position of BT (ly_BT = 0)
    omega_VSP1 = 0.199; % turning rate of VSP1
    omega_VSP2 = 0.235; % turning rate of VSP2
    f_VSP1_max = 0.477; % max thrust from VSP1 (with omega_VSP1 = 0.199)
    f_VSP2_max = 0.433; % max thrust from VSP2 (with omega_VSP2 = 0.235)
    f_BT_max = 0.27;    % max thrust from BT (min. of +/- y-dir.)
    %%%%%%%%%%%%%%%%%%%%%%%%%%%
    
    alpha = zeros(2,1); % VSP angles
    omega = zeros(2,1); % VSP angular rates
    u = zeros(3,1);     % Mapped force input to the VSPs and BT

    % Extended thrust configuration matrix
    B = [  1       0       1       0       0;
           0       1       0       1       1;
         ly_VSP -lx_VSP -ly_VSP -lx_VSP  lx_BT];

    B_dagger = pinv(B);

    f_cmd = B_dagger*tau;

    u_cmd = [norm(f_cmd(1:2))/f_VSP1_max;
             norm(f_cmd(3:4))/f_VSP2_max;
             f_cmd(5)/f_BT_max];
     
    % VSP input is linear from 0.3 to 1
    u_cmd(1:2) = 0.7.*u_cmd(1:2) + 0.3;
    
    % Saturation: Limit input signal to VSPs to be within (0,1) 
    if u_cmd(1) < 0
        u_cmd(1) = 0;
    elseif u_cmd(1) > 1
        u_cmd(1) = 1;
    else
        u_cmd(1) = u_cmd(1);
    end

    if u_cmd(2) < 0
        u_cmd(2) = 0;
    elseif u_cmd(2) > 1
        u_cmd(2) = 1;
    else
        u_cmd(2) = u_cmd(2);
    end

    % Saturation: Limit input signal to BT to be within (-1,1)
    if u_cmd(3) < -1
        u_cmd(3) = -1;
    elseif u_cmd(3) > 1
        u_cmd(3) = 1;
    else
        u_cmd(3) = u_cmd(3);
    end
    
    alpha(1) = atan2(f_cmd(2),f_cmd(1)); % angle of VSP1
    alpha(2) = atan2(f_cmd(4),f_cmd(3)); % angle of VSP2
    
    omega(1) = omega_VSP1;
    omega(2) = omega_VSP2;

    u(1) = u_cmd(3); % input to BT
    u(2) = u_cmd(1); % input to VSP1
    u(3) = u_cmd(2); % input to VSP2
    
    % Thrust configuration matrix (rotable actuators)
    B2 = [cos(alpha(1)) cos(alpha(2)) 0; 
          sin(alpha(1)) sin(alpha(2)) 1; 
          (-ly_VSP*cos(alpha(1))-lx_VSP*sin(alpha(1)))...
          (ly_VSP*cos(alpha(2))-lx_VSP*sin(alpha(2))) lx_BT];
    % Actual force output  
    tau_out = B2*(u_cmd.*[f_BT_max,f_VSP1_max,f_VSP2_max]');
    
end