% Hybrid A* path search based on code for the A* algorithm by: 
% Premakumar, P.(2016): A* (A Star) search for path planning tutorial. 
% Version 1.2.0.1. Available at: https://se.mathworks.com/matlabcentral/
% fileexchange/26248-a-a-star-search-for-path-planning-tutorial (accessed:
% 01.11.2019)

% Cost function based on information provided by the BINN guidance model
% Penalty on heading changes equal to the heading in radians
function [psi_p_new, pos_new, idx_new, optimal_path] = ...
    hybridAstarBINN_kinematics(Z, p0, psi0, U, r_max, Nx, Ny, dX, dY,...
    Lx, Ly, x, y, plot_search, plot_path, heuristics)

    %% Hybrid A* parameters
    rot = [-.7*r_max,1e-8,.7*r_max]'; %[rad/s] Turning rates
    dPsi = 10 *pi/180; %[rad] Resolution of the heading angle (10deg)
    posAccuracy = max(dX,dY); %[m] Radius of acceptance
    %headingAccuracy = dPsi; %[rad] Accepted deviation in heading %%NB! No specified desired heading at the target%%

    %% Obstacles and Targets
    % Target
    [~,idx_t] = min(Z(:)); % Find target based on neural activity
    targetCell = zeros(1,2); % Cell containing the target
    [targetCell(1),targetCell(2)] = findCell(x(idx_t),y(idx_t),dX,dY);
    
    % Obstacles
    idx_o = find(Z>0);  % Find obstacles based on neural activity
    No = length(idx_o); % Number of obstacles
    obstacleCell = zeros(No,2);
    obstacleCell(:,1) = floor(idx_o/Ny)+1; % Obstructed cells
    obstacleCell(:,2) = idx_o-Ny*(obstacleCell(:,1)-1);

    %% Lists used for the A* algorithm
    % Open List Structure
    % --------------------------------------------------------------------
    % |is on list 1/0|xCell|yCell|psi|xPos|yPos|idx_parent|h(n)|g(n)|f(n)|
    % --------------------------------------------------------------------
    OPEN = zeros(10000,10); % Preallocate memory
    % Closed List Structure
    % -----------------------------------
    % |xCell|yCell|noClosedPsi|closedPsi|
    % -----------------------------------
    CLOSED = zeros(10000,3+floor(2*pi/dPsi)+1); % Preallocate memory
    
    CLOSED(1:No,:) = [obstacleCell,... % Insert all obstructed cells (closed for all headings)
        repmat([floor(2*pi/dPsi)+1,(-pi:dPsi:pi)],No,1)];
    CLOSED_count = No; % # nodes in closed list

    % Initial state
    xPos = p0(1); yPos = p0(2); % Initial node
    [xCell,yCell] = findCell(xPos, yPos, dX, dY); % Intial cell
    psi = roundValue(psi0,dPsi); % [rad] Initial heading (discrete)
    path_cost = 0; % Cost along expanded path
    if heuristics % Cost to target
        % Euclidean distance adjusted for neural activity
        gn = 0.5*(Z(targetCell(2),targetCell(1))+1)*... 
            distance(xPos, yPos, x(idx_t), y(idx_t));
    else
        % Euclidean distance
        gn = distance(xPos, yPos, x(idx_t), y(idx_t)); 
    end
    OPEN_count = 1; % # nodes in open list
    OPEN(OPEN_count,:) = [1,xCell,yCell,psi,xPos,yPos,...
        OPEN_count,path_cost,gn,gn]; % Insert to open list
    OPEN(OPEN_count,1) = 0; % Remove from open list    
    CLOSED_count = CLOSED_count + 1; 
    CLOSED(CLOSED_count,1:4) = [xCell,yCell,1,psi]; % Insert to closed list
    
    if plot_search % Initialization of search tree plot
        figure(21); clf; hold on;      %,'visible','off'
        %plot(y,x,'.','Color',0.8*[1 1 1],'Linewidth',2); % Nodes
        plot(y(idx_t),x(idx_t),'p','Color',[0.6350 0.0780 0.1840],... 
            'MarkerFaceColor',[0.6350 0.0780 0.1840]); % Target 
        plot(y(idx_o),x(idx_o),'ks','MarkerFaceColor','k'); %Obstacles
        [yGrid,xGrid] = meshgrid(0:dY:Ly,0:dX:Lx); % Grid
        plot(yGrid,xGrid,':','Color',0.8*[1 1 1]);
        plot(yGrid',xGrid',':','Color',0.8*[1 1 1]);
        xlabel('y [m]'); ylabel('x [m]');
        axis equal tight, box on
        hold off;
    end
    
    %% Hybrid A* algorithm
    idx_OPEN_min = OPEN_count;
    noPath = 0; % there exists no path = false
    while( norm([x(idx_t)-xPos,y(idx_t)-yPos]) > posAccuracy && ~noPath )

        % Expanded Array Structure
        % ------------------------------------------
        % |xCell|yCell|psi|xPos|yPos|h(n)|g(n)|f(n)|
        % ------------------------------------------ 
        exp_array = expand_array2(psi, xPos, yPos, xCell, yCell,...
            path_cost, x(idx_t), y(idx_t), CLOSED(1:CLOSED_count,:),...
            Lx, Ly, dX, dY, rot, U, dPsi, Z, plot_search, heuristics);
        exp_num = size(exp_array,1);
        for i = 1:exp_num
            flag = 0;
            for j = 1:OPEN_count
                if(abs(exp_array(i,4) - OPEN(j,5)) < 1e-5 ...
                        && abs(exp_array(i,5) - OPEN(j,6)) < 1e-5)
                    % Node is already expanded from another direction
                    flag = 1;
                elseif(exp_array(i,1) == OPEN(j,2) && exp_array(i,2)...
                      == OPEN(j,3) && abs(exp_array(i,3)-OPEN(j,4)) < 1e-5)
                    % Cell is already approached with this angle (keep the cheapest)
                    OPEN(j,10) = min(OPEN(j,10),exp_array(i,8));
                    if abs(OPEN(j,10)-exp_array(i,8)) < 1e-5
                        % Update parent, gn, hn
                        OPEN(j,5:10) = [exp_array(i,4:5),idx_OPEN_min,...
                            exp_array(i,6:8)];
                    end % End of check for minimum fn
                    flag = 1;
                end 
            end % End of checking if child node already exists in open list 
            if ~flag
                OPEN_count = OPEN_count + 1;
                OPEN(OPEN_count,:) = [1,exp_array(i,1:5),idx_OPEN_min,...
                    exp_array(i,6:8)];
            end % End of inserting new element into the open list        
        end % End of i loop

        % Find node with the smallest total cost, fn
        idx_OPEN_min = min_fn(OPEN,OPEN_count,x(idx_t),y(idx_t),posAccuracy);
        if(idx_OPEN_min ~= -1) % If there exists a feasible path:
            % Select node with the smallest total cost, fn, or the target
            xCell = OPEN(idx_OPEN_min,2);
            yCell = OPEN(idx_OPEN_min,3);
            psi = OPEN(idx_OPEN_min,4);
            xPos = OPEN(idx_OPEN_min,5);
            yPos = OPEN(idx_OPEN_min,6);
            % Update the cost of reaching the parent node
            path_cost = OPEN(idx_OPEN_min,8);
            % Move the node to the CLOSED list
            [inCLOSED,idx_CLOSED] = ismember([xCell yCell],CLOSED(:,1:2),'rows');
            if inCLOSED
                CLOSED(idx_CLOSED,3) = CLOSED(idx_CLOSED,3)+1;
                CLOSED(idx_CLOSED,CLOSED(idx_CLOSED,3)+3) = psi;
            else
                CLOSED_count = CLOSED_count + 1;
                CLOSED(CLOSED_count,1:4) = [xCell yCell 1 psi];
            end
            OPEN(idx_OPEN_min,1) = 0;
        else % There does not exist a path to the target
            noPath = 1; % Exits the loop!
        end 

    end

    if(noPath == true) % No feasible path: Change operation mode to DP
        fprintf('No feasible path to the target: Operation mode is changed to DP.\n');
        pos_new = p0';
        [xCell,yCell] = findCell(p0(1),p0(2),dX,dY);
        idx_new = (xCell-1)*Ny + yCell;
        psi_p_new = psi0;
        optimal_path = p0;
        return;
    else % Traverse the open list and extract the path  
        optimal_path = zeros(Nx*Ny,2);
        i = 1;
        % Adds target to the path instead of the last expanded node
        optimal_path(i,:) = [x(idx_t) y(idx_t)];
        parent_idx = OPEN(idx_OPEN_min,7);
        while(OPEN(parent_idx,5) ~= p0(1) || OPEN(parent_idx,6) ~= p0(2))
            i = i + 1;
            optimal_path(i,:) = OPEN(parent_idx,5:6);
            parent_idx = OPEN(parent_idx,7); % Grandparent node
        end
        i = i + 1;
        optimal_path(i,:) = p0;
        optimal_path = optimal_path(1:i,:);
    end 
    [nWP,~] = size(optimal_path);
    % Return 2nd next WP if sufficiently close (Equally feasible)
    if nWP > 2 && (norm(diff(optimal_path(i-2:i-1,:))) < min(dX,dY)/3 ...
            || norm(diff(optimal_path(i-1:i,:))) < min(dX,dY)/3) 
        pos_new = optimal_path(i-2,:);
    else
        pos_new = optimal_path(i-1,:);
    end
    [xCell,yCell] = findCell(pos_new(1),pos_new(2),dX,dY);
    idx_new = (xCell-1)*Ny + yCell;
    psi_p_new = atan2(pos_new(2)-p0(2),pos_new(1)-p0(1));
    
    if plot_search
        figure(21); hold on;
        plot(optimal_path(:,2),optimal_path(:,1),'.-',... % Optimal path
            'Color',[0.4660 0.6740 0.1880],'MarkerSize',12,'LineWidth',1.5);
        hold off;
    end

    if plot_path % Plot path
        figure(22); clf; hold on;       
        %plot(y,x,'.','Color',0.8*[1 1 1],'Linewidth',2); % Nodes
        plot(y(idx_t),x(idx_t),'p','Color',[0.6350 0.0780 0.1840],... 
            'MarkerFaceColor',[0.6350 0.0780 0.1840]); % Target
        plot(y(idx_o),x(idx_o),'ks','MarkerFaceColor','k'); % Obstacles
        [yGrid,xGrid] = meshgrid(0:dY:Ly,0:dX:Lx); % Grid
        plot(yGrid,xGrid,':','Color',0.8*[1 1 1]);
        plot(yGrid',xGrid',':','Color',0.8*[1 1 1]); 
        plot(optimal_path(:,2),optimal_path(:,1),'.-',... % Optimal path
            'Color',[0.4660 0.6740 0.1880],'LineWidth',1,'MarkerSize',10); 
        xlabel('y [m]'); ylabel('x [m]');
        axis equal tight, box on
        hold off;
    end
    
    
    %% Functions 
    function dist = distance(x1,y1,x2,y2)

        % This function calculates the distance between 
        % any two cartesian coordinates.
        %
        % Copyright 2009-2010 The MathWorks, Inc.

        dist = sqrt((x1-x2)^2 + (y1-y2)^2);

    end

    function exp_array = expand_array2(psi, xPos, yPos, xCell, yCell, hn,...
            xTarget, yTarget, CLOSED, Lx, Ly, dX, dY, turnRates, U, dPsi,...
            Z, plot_search, heuristics) %(psi,xPos,yPos,hn,xTarget,yTarget,CLOSED,Lx,Ly,dX,dY,turnRates,dPsi,Z)%

        % Function generating an expanded array:
        % Taking a node as input, it returns the expanded list of 
        % successors, with the calculated total cost (fn). The criteria 
        % being none of the successors are on the CLOSED list.
        % 
        % Expanded Array Structure
        % ------------------------------------------
        % |xCell|yCell|psi|xPos|yPos|h(n)|g(n)|f(n)|
        % ------------------------------------------
        %
        % Copyright 2009-2010 The MathWorks, Inc.
        % Modified: Caroline Fleischer, 24.04.2020

%         [xCellCurrent,yCellCurrent] = findCell(xPos,yPos,dX,dY);
         exp_array = nan(length(turnRates),8);
         exp_count = 0;
         for k = 1:length(turnRates) % Generate child nodes
             insert = true;
             r = turnRates(k);     
             phi0 = psi + pi/2; % [rad] angle normal to the heading
             R = U/r; % [m] turn radius
             c0 = [xPos,yPos]' + R*[cos(phi0); sin(phi0)]; % center of turn circle
             cellNext = [xCell,yCell];
             theta = psi - pi/2; % [rad] central angle
             dtheta = dX/(12*abs(R)); % increment of the central angle
             n = 1; % iteration counter
             in = false; % true = cell is in the closed list
             idxC = -1; % index in closed list
             c = [[xPos,yPos]', zeros(2,50)]; 
             while cellNext(1) == xCell && cellNext(2) == yCell % i <= N-1   
                 n = n+1;
                 theta = theta + sign(r)*dtheta;
                 c(:,n) = c0 + R*[cos(theta); sin(theta)];
                 [cellNext(1),cellNext(2)] = findCell(c(1,n),c(2,n),dX,dY);
                 % Check whether an obstructed cell is encountered
                 [in,idxC] = ismember(cellNext,CLOSED(:,1:2),'rows');
                 if in && CLOSED(idxC,3) >= floor(2*pi/dPsi)+1 
                    insert = false;
                    break;
                 end
                 if n > 50
                     error('Infinite loop: Turn radius is too small to get out of the current cell (\ttheta= %.2f[deg]).\n',theta*180/pi);
                 end
             end
             c = c(:,1:n);
             if plot_search
                 set(0,'CurrentFigure',figure(21));
                 line(c(2,:),c(1,:),'Color',1.2*[0.4660 0.6740 0.1880]);
             end

             psiEnd = roundValue(ssa(theta+pi/2,'rad'),dPsi); % discrete heading
             theta = theta-psi+pi/2; % final central angle
             l = R*theta; % traveled distance
             
             % Check if cell is closed for psiNew
             if in && ismember(psiEnd,CLOSED(idxC,4:3+CLOSED(idxC,3)))
                 insert = false;  
             end
 
             if insert && (c(1,end) >= 0 && c(1,end) <= Lx) && (c(2,end) >= 0 && c(2,end) <= Ly)
                 exp_count = exp_count + 1; 
                 % The cell where the node is located
                 exp_array(exp_count,1) = cellNext(1);
                 exp_array(exp_count,2) = cellNext(2);
                 % Path angle and node position
                 exp_array(exp_count,3) = psiEnd;
                 exp_array(exp_count,4) = c(1,end);
                 exp_array(exp_count,5) = c(2,end);
                 % Cost of travelling to node, hn
                 if heuristics 
                     % Traveled distance adjusted for neural activity
                     exp_array(exp_count,6) = hn + ...
                        0.5*(Z(cellNext(2),cellNext(1))+1)*l;%distance(xPos,yPos,c(1,end),c(2,end)); 
                 else
                     % Traveled distance
                     exp_array(exp_count,6) = hn + l;%distance(xPos,yPos,c(1,end),c(2,end));
                 end
                 % Cost from node to goal, gn (penalty on direction change and heading error)
                 if heuristics
                     % Heading error relative to progression straight towards the target
                     psi_e = abs(atan2(yTarget-c(2,end),xTarget-c(1,end))-psiEnd);
                     % Target cell
                     [xTCell,yTCell] = findCell(xTarget,yTarget,dX,dY);
                     % Euclidean distance adjusted for neural activity
                     exp_array(exp_count,7) = abs(r) + psi_e + ...
                         0.5*(Z(yTCell,xTCell)+1)*...
                         distance(xTarget,yTarget,c(1,end),c(2,end));
                 else
                     % Euclidean distance
                     exp_array(exp_count,7) = distance(xTarget,yTarget,c(1,end),c(2,end));
                 end
                 % Total cost, fn
                 exp_array(exp_count,8) = exp_array(exp_count,6)...
                     + exp_array(exp_count,7); 
             end % End of node within workspace     
         end % End of k for loop
         exp_array = exp_array(1:exp_count,:);
    
    end

    function [xCell,yCell] = findCell(x,y,dX,dY)

        % Function returning the corresponding cell coordinates to a position 
        % within the grid 
        %
        % Caroline Fleischer, 03.03.2020

        xCell = ceil(x/dX);
        yCell = ceil(y/dY);
    
    end

    function idx_min = min_fn(OPEN,OPEN_count,xTarget,yTarget,posAccuracy)

        % Function that returns the node with the smallest total 
        % cost, fn, or the goal node. It takes the OPEN List as input.
        %
        % Copyright 2009-2010 The MathWorks, Inc.

        temp_array = nan(OPEN_count,11);
        k = 1;
        % Iterate through all nodes in the open list
        for n = 1:OPEN_count
            if (OPEN(n,1) == 1)
                temp_array(k,:) = [OPEN(n,:) n];
                % Return the goal node if it is among the successors
                if norm([xTarget-OPEN(n,5), yTarget-OPEN(n,6)]) <= posAccuracy
                    idx_min = n; 
                    return;
                end 
                k = k + 1;
            end
        end 
        temp_array = temp_array(1:k-1,:);
        % Return the open list index of the node with the smallest f
        if size(temp_array ~= 0)
            [~,temp_min] = min(temp_array(:,10)); % Index in temp_array
            idx_min = temp_array(temp_min,11); % Index in the open list
        else
            idx_min = -1; % temp_array is empty (i.e. no feasible path)
        end

    end

    function roundedValue = roundValue(value,step)

        % Function that returns a value rounded to a certain resolution
        %
        % Adopted from Nordeus, E. (2019): Self-driving vehicle. Available 
        % at: https://github.com/Habrador/Self-driving-vehicle/blob/master...
        % /Self-driving%20vehicle%20Unity/Assets/Scripts/Pathfinding...
        % /Utility/HelpStuff.cs (accessed: 03.03.2020)

        roundedValue = round(value/step)*step;

    end

end