clc,clear

%% Solve MIP in AMPL to obtain a path
addpath(genpath('/AMPL'))
setUp

% Create an AMPL instance
ampl = AMPL('/AMPL');

% Initialisation
basef = fileparts(which('mip'));
modeldirectory = fullfile(basef, 'AMPL_model');

% Solver
ampl.setOption('solver', 'knitro');

% Load from file the ampl model
ampl.read([modeldirectory  '/' 'mip2.mod']);
ampl.readData([modeldirectory  '/' 'mip.dat']);

% Solve
ampl.solve

% Get cost
cost = ampl.getObjective('cost'); cost = cost.value();

% Get data
df = ampl.getData('x','y'); df2 = ampl.getData('z');
path = [df.getColumnAsDoubles('x'),df.getColumnAsDoubles('y')];
z_chi = df2.getColumnAsDoubles('z');

% Close the AMPL object
ampl.close(); 

%% Workspace parameters
Lx = 20;                    % Length in x-direction (north)
Ly = Lx;                    % Length in y-direction (east)

Nx = 4; dX = Lx/Nx;         % Grid partitioning (regular)
Ny = Nx; dY = Ly/Ny;
N  = Nx*Ny;                 % Total number of grid elements/neurons

[X,Y] = meshgrid(dX/2:dX:Lx,dY/2:dY:Ly);                 
x     = reshape(X,[N,1]);
y     = reshape(Y,[N,1]);

% Target and obstacles
TargetIdx = [Nx,1];
idx_t = (TargetIdx(1)-1)*Ny + TargetIdx(2);

ObstacleIdx = [2,1; 2,2; 3,1; 3,2];
idx_o = (ObstacleIdx(:,1)-1)*Ny + ObstacleIdx(:,2);

%% Plotting
set(groot,'defaulttextinterpreter','latex',...
    'defaultAxesTickLabelInterpreter','latex',...
    'defaultLegendInterpreter','latex',...
    'defaultTextboxshapeInterpreter','latex'); 

figure(1); clf; hold on;
plot(y,x,'.','Color',0.8*[1 1 1],'Markersize',12); % Neurons
[yGrid,xGrid] = meshgrid(0:dY:Ly,0:dX:Lx); % Grid
plot(yGrid,xGrid,':','Color',0.8*[1 1 1]);
plot(yGrid',xGrid',':','Color',0.8*[1 1 1]); xlabel('y');
plot(y(idx_t),x(idx_t),'p','Color',[0.6350 0.0780 0.1840],'MarkerFaceColor',[0.6350 0.0780 0.1840],'Markersize',12); % Targets
plot(y(idx_o),x(idx_o),'ks','MarkerFaceColor','k','Markersize',10); % Obstacles
plot(path(:,2),path(:,1),'.-','Color',[0 0.4470 0.7410],'LineWidth',1.5,'Markersize',15); % Path
lz_chi = length(z_chi)/2; % length of unit velocity vector
psi_p = length(z_chi)/2; % path-tangential angle
for i = 1:length(z_chi)/2
    psi_p(i) = atan2(z_chi(2*i),z_chi(2*i-1));
    R_p = [cos(psi_p(i)), sin(psi_p(i)); -sin(psi_p(i)), cos(psi_p(i))];
    arrowhead = R_p'*[-.3, 0, -.3, -.3; -.1, 0, .1, -.1];
    plot(path(i,2)+z_chi(2*i)+arrowhead(2,:),path(i,1)+z_chi(2*i-1)+arrowhead(1,:),...
        'Color',[0.8500 0.3250 0.0980],'LineWidth',2); 
    plot(path(i,2)+[0,z_chi(2*i)]',path(i,1)+[0,z_chi(2*i-1)]',...
        'Color',[0.8500 0.3250 0.0980],'LineWidth',2); hold on;
    lz_chi(i) = sqrt(z_chi(2*i-1)^2+z_chi(2*i)^2);
    if mod(i,2)
    text(path(i,2)+.5,path(i,1)+1,'$|z^\chi|=$' + "" + num2str(lz_chi(i),'%.2f\n'))
    end
end
cost_str = strcat('cost:'," ",num2str(cost,'%.1e\n'));
annotation('textbox',[.64 .62 .3 .3],'String',cost_str,'FitBoxToText','on');
xlabel('y')
ylabel('x')
axis equal tight, box on
hold off;