%Initialize and run simulation of CSE1 based on Fossen's kinematic 
%and kinetic vehicle model. The guidance system comprises a hybrid A* path 
%planner and LOS path-following law. 

clc, clear, close all
%clear 'hybridAstarBINN.m' 'hybridAstarBINN_kinematics.m' 
plot_prelims

%% User preferences for path-planning
plot_search = 0; % plot hybrid A* search tree: 1 = on, 0 = off
plot_path   = 0; % plot hybrid A* path at each iteration: 1 = on, 0 = off

expansion  = 0; % 0 = straight-line, 1 = kinematics-based
heuristics = 1; % 0 = Euclidean distance, 1 = adjusted for neural activity

%% Angle conversion factors
rad2deg = 180/pi; % radians to degrees
deg2rad = pi/180; % degrees to radians

%% Simulation parameters
h = 0.01;     %[s] time step
Tstop = 1500; %[s] simulation time

%% CSE1 data
L = 1.105; %[m] Vehicle length

%% Path-planning parameters
Uref  = 0.2;       %[m/s] Desired reference speed  %% NB: Vessel is 1:50 %%
Rmin  = 3/2 *L;    %[m] Minimum turn radius
r_max = Uref/Rmin; %[rad/s] Maximum rotation rate

% Workspace parameters
Lx = 42; % Length in x-direction (north)
Ly = Lx; % Length in y-direction (east)

Nx = 14; dX = Lx/Nx; % Grid partitioning (regular)
Ny = Nx; dY = Ly/Ny;
N  = Nx*Ny;          % Total number of grid cells/neurons

[X,Y] = meshgrid(dX/2:dX:Lx,dY/2:dY:Ly);                 
x     = reshape(X,[N,1]);
y     = reshape(Y,[N,1]);
r0    = min(dX,dY)/8; %[m] Radius of acceptance (defining when the waypoint is regarded reached by the vessel)

% Neural Network parameters
E    = 10;         % Constant for external input
d0   = max(dX,dY); % Radius of the receptive field R (defining the neighboring neurons)
a    = 10;         % Passive decay rate (shunting eq.)
bu   = 1.0;        % Upper bound of neural activity (shunting eq.)            
bl   = 1.0;        % Lower bound of neural activity (shunting eq.)
cw   = 1;          % Constant defining connection weight between neurons
e_ss = 1e-10;      % Condition for steady-state neural activity 

% Matrices of neighbouring neurons A(i,j) and connection weights W(i,j)
[A,W] = neighborNeurons(x,y,N,d0,cw);

% Parameters for COLAV compliant with COLREGs
rd   = 20;                  % Radius of the detection region
sd_L = 10*L;                % large safe distance
sd_M = 5*L;                 % medium safe distance
sd_S = 2*L;                 % small safe distance

% Initial ASV position and heading
idx0 = Ny;                  % Initial position index (lower right corner)
psi0 = 0*deg2rad;           %[rad] Initial heading (towards north)
eta0 = [x(idx0),y(idx0),psi0]'; %[m m rad]' Initial state

% Target
TargetIdx = [Nx, 1];        % Target index (upper left corner)
[Nt,~] = size(TargetIdx);   % Number of targets
idx_t = (TargetIdx(:,1)-1).*Ny ...
    + TargetIdx(:,2);       % Indices of targets in vector grid          

% Static obstacles
StaticObIdx = [];%[6,5; 7,3; 7,4; 7,5; 7,6; 8,4; 8,5; 8,6; 8,7; 12,6; 13,6; 14,6];%[1,1; 2,1; 3,1; 1,2; 2,2; 6,6; 7,7; 8,8];
if ~isempty(StaticObIdx)
    idx_so = (StaticObIdx(:,1)-1).*Ny ...
        + StaticObIdx(:,2); % Indices of static obstacles in vector grid 
else
    idx_so = 0;
end

% Moving obstacle
r_obj   = 1;                              % [m] radius of object
p0_obj  = [38;33];                         % [m,m]' initial position                                   
U_obj   = 0.9*Uref;                       % [m/s] velocity
d0_obj  = -4*pi/5;                         % [rad] initial direction 
theta   = linspace(0, 2*pi);
obj_def = r_obj*[cos(theta); sin(theta)]; % object definition (circle)

%% Motion control parameters
Delta = L;   % LOS lookahead distance
Kp    = 0.5; % Proportional gain (course controller)

%% Run simulation
sim('LOS_sim')

%% KPIs of the path planner
% Average execution time for finding WPs (omit start-up and termination time)
t_plan_avg = sum(t_plan(2:end-1))/(length(t_plan)-2); 
% Path length
WP_unique = unique(WP,'rows');
path_length = sum(sqrt(sum(diff([eta0(1:2)'; WP_unique]).^2,2)));
% Accumulated turn
psi_acc = sum(abs(atan2(diff([eta0(2); WP_unique(:,2)]),...
    diff([eta0(1); WP_unique(:,1)]))));
psi_acc_asv = sum(abs(diff(eta(:,3))));

%% Plotting
figure(1); clf; hold on
plot([eta0(2);WP(:,2)],[eta0(1);WP(:,1)],'.-','MarkerSize',12,'Color',green,'DisplayName','waypoints') % Waypoints
plot(eta_obj(:,2), eta_obj(:,1),'-.','Color',orange,'DisplayName','object tracks') % Moving object
plot(eta(:,2), eta(:,1),'-k','DisplayName','vessel tracks') % Vessel tracks
if idx_so ~= 0
    plot(y(idx_so),x(idx_so),'ks','MarkerFaceColor','k','DisplayName','static obstacles'); % Static obstacles
end
plot(y(idx_t),x(idx_t),'p','Color',red,'MarkerFaceColor',red,'DisplayName','target'); % Target
[yGrid,xGrid] = meshgrid(0:dY:Ly,0:dX:Lx); % Grid
plot(yGrid,xGrid,':','Color',gray,'HandleVisibility','off');
plot(yGrid',xGrid',':','Color',gray,'HandleVisibility','off');
axis([0 Ly 0 Lx]), axis equal tight manual
for i = 1:Tstop/(50*h):length(eta)  % Plot vessel
    R0 = [cos(eta(i,3)) -sin(eta(i,3)); 
          sin(eta(i,3))  cos(eta(i,3))];
    ship = R0*[6*L/9 5*L/9 3*L/9 -5*L/9 -5*L/9 3*L/9 5*L/9 6*L/9; 0 1*L/9 2*L/9 2*L/9 -2*L/9 -2*L/9 -1*L/9 0];
    plot(eta(i,2)+ship(2,:), eta(i,1)+ship(1,:),'k','HandleVisibility','off')
end
%Rd = rd * [cos(theta); sin(theta)] + eta(end,1:2)'; % Current detection region
%fill(Rd(2,:),Rd(1,:),'b','LineStyle','none','FaceAlpha',.05);
fill(obj_def(2,:)+eta_obj(end,2),obj_def(1,:)+eta_obj(end,1),... % Current object position
    orange,'FaceAlpha',.2,'EdgeColor',orange,'HandleVisibility','off');
legend('show','Location','northeast') %'static obstacles'
%title('Pragmatic B\''ezier path')
xlabel('y [m]'); xticks(dY:2*dY:Ly)
ylabel('x [m]'); yticks(dX:2*dX:Lx)
% % Zoomed subplot: Create a new pair of axes inside current figure
% axes('position',[.71 .49 .2 .2]) %[left bottom width height]
% box on % put box around new pair of axes
% idxOfInterest = (eta(:,2) > 32) & (eta(:,2) < 35); %(eta(:,2) > 28) & (eta(:,2) < 31);%
% plot(eta(idxOfInterest,2),eta(idxOfInterest,1),'-k'); hold on 
% %plot(etad(idxOfInterest,2),etad(idxOfInterest,1),'-','Color',green)
% plot(WP(idxOfInterest,2),WP(idxOfInterest,1),'-.','MarkerSize',12,'Color',green)
% axis equal tight
hold off


%% Functions 
function [A,W] = neighborNeurons(x,y,N,d0,mu)
    % Determining neighbouring neurons A(i,j) in the BINN
    % and their connection weights W(i,j)
    % Input: 
    % x,y - coordinates of neurons in the network
    % N   - total number of neurons in the network
    % mu  - constant defining the connection between neurons
    A = zeros(N,N);     % Adjacency matrix
    W = zeros(N,N);     % Connection weights between neurons
    for i = 1:N
        for j = 1:N
            dist = max(abs(x(j)-x(i)),abs(y(j)-y(i)));
    %        dist = sqrt((x(j)-x(i))^2+(y(j)-y(i))^2);
            if (i ~= j) && (dist <= d0)
                A(i,j) = 1;
                W(i,j) = mu/dist;
            end
        end
    end
end
